#TODO - collision detection - switch to ODE
#TODO - line showing height
#TODO - alt-tab fixing. Use mouse.get_focused - doesn't seem to work


import pygame
import sys
import math
from data import filepath

import random

import ode

from BkGlobals import *
import BtOptionsMenu
import BtDifficultyMenu

def setupScreen(width, height, bWindowed):
	global SCREEN_X
	global SCREEN_Y
	global MAX_BLOCK_SIZE_X
	global MAX_BLOCK_SIZE_Y
	global BORDER_WIDTH_X
	global BORDER_WIDTH_Y
	global GAME_SURFACE_POS
	global GAME_SURFACE_SIZE
	global NEXT_BLOCK_SURFACE_POS
	global NEXT_BLOCK_SURFACE_SIZE
	global TEXT_BORDER_X
	global TEXT_BORDER_Y
	
	if not width or not height:
		dispModes = pygame.display.list_modes()
		dispModes.sort(reverse = True)
	
		width = dispModes[0][0]
		height = dispModes[0][1]
		
	pygame.display.set_caption("PyBuckle")
	
	if bWindowed:
		pygame.display.set_mode((width, height))
	else:
		pygame.display.set_mode((width, height), pygame.FULLSCREEN)
	
	BORDER_WIDTH_X = BORDER_WIDTH_X * width / SCREEN_X
	BORDER_WIDTH_Y = BORDER_WIDTH_Y * height / SCREEN_Y
		
	MAX_BLOCK_SIZE_X = MAX_BLOCK_SIZE_X * width / SCREEN_X
	MAX_BLOCK_SIZE_Y = MAX_BLOCK_SIZE_Y * height / SCREEN_Y
					
	GAME_SURFACE_POS = (BORDER_WIDTH_X, BORDER_WIDTH_Y)
	GAME_SURFACE_SIZE = ((3 * width / 4) - BORDER_WIDTH_X, height - (2 * BORDER_WIDTH_Y))
	NEXT_BLOCK_SURFACE_POS = (GAME_SURFACE_POS[0] + GAME_SURFACE_SIZE[0], BORDER_WIDTH_Y)
	NEXT_BLOCK_SURFACE_SIZE = width - GAME_SURFACE_SIZE[0] - 2 * BORDER_WIDTH_X
	
	TEXT_BORDER_X = TEXT_BORDER_X * width / SCREEN_X
	TEXT_BORDER_Y = TEXT_BORDER_Y * height / SCREEN_Y

	
	SCREEN_X = width
	SCREEN_Y = height

			
def runGame(bFirstCall = True, width = 800, height = 600, bWindowed = False):
	if bFirstCall:
		setupScreen(width, height, bWindowed)
	pygame.display.get_surface().fill(BG_COLOUR)	
	
	world = ode.World()
	world.setGravity((0, -150 ,0))
	world.setERP(0.2)
	world.setContactSurfaceLayer(0.01)
	#world.setContactMaxCorrectingVel(100)
	world.setCFM(10E-10)	
	world.setQuickStepNumIterations(100)
	space = ode.Space()
	
	
	currentBlock = Block(world, space)
	nextBlock = Block(world, space)
	
	currentBlock.setSelected(True)
	
	gameArea = GameArea(world, space)
	nextBlockArea = NextBlockArea()
	optionsArea = OptionsArea()
	infoArea = InfoArea()
	
	nextBlockArea.update(nextBlock)
	
	frameRateClock = pygame.time.Clock()
	frameRateClock.tick()
		
	pygame.display.flip()	
		
	bPaused = False
	iTick = 0
	iTime = 0
	iVictoryTimer = WIN_TIME_MEDIUM * 1000
	iOldTime = 0
	updateRectList = []
	numBlocks = NUM_BLOCKS_MEDIUM
	#bOldFocused = True
	
	iBlockNoSpawnTimer = 0
		
	while 1:	
		# if not pygame.mouse.get_focused() or not pygame.display.get_active():
			# bOldFocused = False
			# continue
		# elif not bOldFocused:
			# pygame.display.flip()
			# bOldFocused = True
			
			
		iBlockNoSpawnTimer = max(0, iBlockNoSpawnTimer - 1)
		oldUpdateRectList = updateRectList
		updateRectList = []
		
		if not B_SANDBOX:
			if int(iVictoryTimer / 1000) <= 0:
				bWin = True
				break
			if START_TIME_MEDIUM - int(iTime / 1000) <= 0:
				bWin = False
				break
			if numBlocks == 0 and iVictoryTimer >= WIN_TIME_MEDIUM:
				bWin = False
				break
			
		if bPaused:
			pygame.mouse.set_visible(True)
			frameRateClock.tick(GFX_TICK_RATE)
			for event in pygame.event.get():
				if event.type == pygame.QUIT:
					sys.exit(1)
				elif event.type == pygame.KEYDOWN:
					if event.dict["key"] == pygame.K_ESCAPE:
						sys.exit(1)	
					elif event.dict["key"] == pygame.K_p or event.dict["key"] == pygame.K_PAUSE:
						bPaused = False		
				elif event.type == pygame.MOUSEMOTION:
					pos = event.dict["pos"]
						
					updateRectList += optionsArea.updateMousePos(pos)	
				elif event.type == pygame.MOUSEBUTTONUP:
					if event.dict["button"] == 1:
						pos = event.dict["pos"]
						
						optionsArea.mouseClick(pos)
						
				elif event.type == EVENT_PAUSE:
					bPaused = False
					
				elif event.type == EVENT_OPTIONMENU:
					BtOptionsMenu.OptionsScreen()
					updateRectList.append(pygame.display.get_surface().get_rect())
					break
					
					
			updateRectList += gameArea.update(0)

					
		else:
			iOldTime = iTime
			iTime += frameRateClock.tick(GFX_TICK_RATE)
			
			for event in pygame.event.get():
				if event.type == pygame.QUIT:
					sys.exit(1)
				elif event.type == pygame.KEYDOWN:
					if event.dict["key"] == pygame.K_ESCAPE:
						sys.exit(1)
					elif event.dict["key"] == pygame.K_p or event.dict["key"] == pygame.K_PAUSE:
						bPaused = True
					elif event.dict["key"] == pygame.K_SPACE:
						pos = pygame.mouse.get_pos()
						
						if numBlocks and not iBlockNoSpawnTimer and gameArea.setBlock(currentBlock, pos):
							numBlocks -= 1
							currentBlock = nextBlock
							currentBlock.setSelected(True)
							nextBlock = Block(world, space)
							updateRectList += nextBlockArea.update(nextBlock)

							iBlockNoSpawnTimer = 3

						
				elif event.type == pygame.MOUSEMOTION:
					pos = event.dict["pos"]
						
					updateRectList += optionsArea.updateMousePos(pos)

					updateRectList 
						
				elif event.type == pygame.MOUSEBUTTONUP:
					if event.dict["button"] == 1:
						pos = event.dict["pos"]
						
						if numBlocks and not iBlockNoSpawnTimer and gameArea.setBlock(currentBlock, pos):
							numBlocks -= 1
							currentBlock = nextBlock
							currentBlock.setSelected(True)
							nextBlock = Block(world, space)
							updateRectList += nextBlockArea.update(nextBlock)
							iBlockNoSpawnTimer = 3
						else:
							optionsArea.mouseClick(pos)
							
				elif event.type == EVENT_PAUSE:
					bPaused = True		

				elif event.type == EVENT_OPTIONMENU:
					BtOptionsMenu.OptionsScreen()
					frameRateClock.tick()
					updateRectList.append(pygame.display.get_surface().get_rect())
					break					
					
					
			gameArea.updateMousePos(pygame.mouse.get_pos(), currentBlock)					
					
			updateRectList += gameArea.update((iTime - iOldTime) / 1000.0)
				
			iVictoryTimer = WIN_TIME_MEDIUM * 1000 - gameArea.checkVictoryConditions(iTime - iOldTime)	
			
			updateRectList += infoArea.update(START_TIME_MEDIUM - int(iTime / 1000), int(iVictoryTimer / 1000), numBlocks)
				
				
			iTick += 1
				
		if updateRectList:
			pygame.display.update(updateRectList + oldUpdateRectList)
			
			
	runEnd(gameArea, oldUpdateRectList, bWin)
	runGame(False)
		
			

def showWinScreen():
	return drawFullScreenText("You win!")

def showLoseScreen():
	return drawFullScreenText("You lose!")
	
def drawFullScreenText(szMessage):
	font = pygame.font.Font(pygame.font.get_default_font(), 100)
	message = font.render(szMessage, True, TEXT_FULL_SCREEN_COLOUR).convert_alpha()

	displaySurf = pygame.display.get_surface()
	
	return [displaySurf.blit(message, (displaySurf.get_rect().center[0] - message.get_width() / 2, displaySurf.get_rect().center[1] - message.get_height() / 2))]
					
def runEnd(gameArea, oldUpdateRectList, bWin):
	iTime = 0	
	updateRectList = oldUpdateRectList
	
	frameRateClock = pygame.time.Clock()
	frameRateClock.tick()
	
	while iTime < END_TIME:
		oldUpdateRectList = updateRectList
		updateRectList = []	
	
		for event in pygame.event.get():
			if event.type == pygame.QUIT:
				sys.exit(1)
			elif event.type == pygame.KEYDOWN:
				if event.dict["key"] == pygame.K_ESCAPE:
					sys.exit(1)	
	

		iOldTime = iTime
		iTime += frameRateClock.tick(GFX_TICK_RATE)	
		
		updateRectList += gameArea.update((iTime - iOldTime) / 1000.0, bNoMouse = True)
		
		if bWin:
			updateRectList += showWinScreen()
		else:
			updateRectList += showLoseScreen()
		
		if updateRectList:
			pygame.display.update(updateRectList + oldUpdateRectList)
			

					
class NextBlockArea:
	def __init__(self):	
		self.size = NEXT_BLOCK_SURFACE_SIZE
		self.baseNextBlockSurf = pygame.Surface((self.size, self.size)).convert()
		self.baseNextBlockSurf.fill(NEXT_BLOCK_BG_COLOUR)
		self.baseNextBlockSurf.fill(DISPLAY_BG_COLOUR, ((self.size - MAX_BLOCK_SIZE_X * 1.25) / 2, (self.size - MAX_BLOCK_SIZE_Y * 1.25) / 2, MAX_BLOCK_SIZE_X * 1.25, MAX_BLOCK_SIZE_Y * 1.25))
		
		pygame.display.get_surface().blit(self.baseNextBlockSurf, NEXT_BLOCK_SURFACE_POS)		
		
		self.nextBlockSurf = self.baseNextBlockSurf.copy().convert()
				
	def update(self, nextBlock):
		self.nextBlockSurf = self.baseNextBlockSurf.copy()

		surf = nextBlock.getSurface()
		rect = surf.get_rect()
		
		self.nextBlockSurf.blit(surf, ((self.size / 2) - rect.center[0], (self.size / 2) - rect.center[1]))
		
		return [pygame.display.get_surface().blit(self.nextBlockSurf, NEXT_BLOCK_SURFACE_POS)]
		
class InfoArea:
	def __init__(self):
		self.width = NEXT_BLOCK_SURFACE_SIZE
		self.height = 13 * (SCREEN_Y - NEXT_BLOCK_SURFACE_SIZE) / 20
		
		self.pos = list(NEXT_BLOCK_SURFACE_POS)
		self.pos[1] += NEXT_BLOCK_SURFACE_SIZE
				
				
		self.imageSize = imageSize = self.height / 5
		self.clockSurf = pygame.transform.scale(pygame.image.load(filepath("clock1.png")), (imageSize, imageSize)).convert_alpha()

		self.blocksLeftSurf = pygame.Surface((imageSize, imageSize)).convert()
		self.blocksLeftSurf.fill((0, 0, 0))
		self.blocksLeftSurf.fill(OLD_BLOCK_COLOUR, (1, 1, imageSize - 2, imageSize - 2))

		
		self.blockClock = self.clockSurf.copy()
		
		smallBlocksSurf = pygame.Surface((self.height / 10, self.height / 10)).convert_alpha()
		smallBlocksSurf.fill((0, 0, 0, 0))		
		smallBlocksSurf.fill((0, 0, 0), (3, 3, (self.height / 10) - 6, (self.height / 10) - 6))
		smallBlocksSurf.fill(OLD_BLOCK_COLOUR, (4, 4, (self.height / 10) - 8, (self.height / 10) - 8))
		pygame.draw.line(smallBlocksSurf, TARGET_LINE_COLOUR, (0, self.height / 30), (self.height / 10, self.height / 30), 2)		
	
		self.blockClock.blit(smallBlocksSurf, (1, 1))
		
		fontSize = 100
		font = pygame.font.Font(pygame.font.get_default_font(), fontSize)
		fontSurfSize = font.size("999")

		while fontSurfSize[0] > self.width + 3 * TEXT_BORDER_X + imageSize or fontSurfSize[1] > self.height / 3:
			fontSize -= 1		
			font = pygame.font.Font(pygame.font.get_default_font(), fontSize)
			fontSurfSize = font.size("999")	
			
		self.font = pygame.font.Font(pygame.font.get_default_font(), fontSize - 1)
		fontSurfSize = font.size("999")
		
		height = fontSurfSize[1]
				
		pygame.display.get_surface().blit(self.clockSurf, (self.pos[0] + TEXT_BORDER_X, self.pos[1] + ((height - imageSize) / 2)))
		pygame.display.get_surface().blit(self.blockClock, (self.pos[0] + TEXT_BORDER_X, self.pos[1] + ((height - imageSize) / 2) + self.height / 3))	
		pygame.display.get_surface().blit(self.blocksLeftSurf, (self.pos[0] + TEXT_BORDER_X, self.pos[1] + ((height - imageSize) / 2) + 2 * self.height / 3))

		self.text1BGSurf = pygame.Surface((self.width - 2 * TEXT_BORDER_X - imageSize, height)).convert()		
		self.text2BGSurf = self.text1BGSurf.copy()
		self.blocksLeftBGSurf = self.text1BGSurf.copy()
		
		self.iOldTimeLeft1 = -1
		self.iOldTimeLeft2 = -1
		self.iOldBlocksLeft = -1

	
	def update(self, iTimeLeft1, iTimeLeft2, iBlocksLeft):	
		updateRects = []
	
		if iTimeLeft1 != self.iOldTimeLeft1:
			if iTimeLeft1 <= 10:
				timeLeft1Text = self.font.render(str(iTimeLeft1), True, TEXT_DANGER_COLOUR, INFO_BG_COLOUR).convert()
			else:
				timeLeft1Text = self.font.render(str(iTimeLeft1), True, TEXT_NORMAL_COLOUR, INFO_BG_COLOUR).convert()

			self.text1BGSurf.fill(INFO_BG_COLOUR)
			self.text1BGSurf.blit(timeLeft1Text, (0, 0))
			
			updateRects.append(pygame.display.get_surface().blit(self.text1BGSurf, (self.pos[0] + 2 * TEXT_BORDER_X + self.imageSize, self.pos[1])))
			
			self.iOldTimeLeft1 = iTimeLeft1
			
			
		if iTimeLeft2 != self.iOldTimeLeft2:	
			if iTimeLeft2 <= 3:
				timeLeft2Text = self.font.render(str(iTimeLeft2), True, TEXT_DANGER_COLOUR, INFO_BG_COLOUR).convert()		
			else:
				timeLeft2Text = self.font.render(str(iTimeLeft2), True, TEXT_NORMAL_COLOUR, INFO_BG_COLOUR).convert()
				
			self.text2BGSurf.fill(INFO_BG_COLOUR)
			self.text2BGSurf.blit(timeLeft2Text, (0, 0))
			
			updateRects.append(pygame.display.get_surface().blit(self.text2BGSurf, (self.pos[0] + 2 * TEXT_BORDER_X + self.imageSize, self.pos[1] + self.height / 3)))
			
			self.iOldTimeLeft2 = iTimeLeft2
			
		if iBlocksLeft != self.iOldBlocksLeft:
			if iBlocksLeft <= 10:
				blocksLeftText = self.font.render(str(iBlocksLeft), True, TEXT_DANGER_COLOUR, INFO_BG_COLOUR).convert()		
			else:
				blocksLeftText = self.font.render(str(iBlocksLeft), True, TEXT_NORMAL_COLOUR, INFO_BG_COLOUR).convert()
				

			self.blocksLeftBGSurf.fill(INFO_BG_COLOUR)
			self.blocksLeftBGSurf.blit(blocksLeftText, (0, 0))
			
			updateRects.append(pygame.display.get_surface().blit(self.blocksLeftBGSurf, (self.pos[0] + 2 * TEXT_BORDER_X + self.imageSize ,self.pos[1] + 2 * self.height / 3)))
					
			self.iOldBlocksLeft = iBlocksLeft
			
		return updateRects
				
class OptionsArea:
	def __init__(self):
		self.width = NEXT_BLOCK_SURFACE_SIZE
		self.height = 2 * (SCREEN_Y - NEXT_BLOCK_SURFACE_SIZE) / 5

		self.pos = list(NEXT_BLOCK_SURFACE_POS)
		self.pos[1] += NEXT_BLOCK_SURFACE_SIZE + (13 * (SCREEN_Y - NEXT_BLOCK_SURFACE_SIZE) / 20)
		
		surf = pygame.Surface((self.width, self.height)).convert()
		surf.fill(OPTIONS_BG_COLOUR)
		
		fontSize = 100
		font = pygame.font.Font(pygame.font.get_default_font(), fontSize)
		fontSurfSize1 = font.size("Options")
		fontSurfSize2 = font.size("Pause")
		fontSurfSize3 = font.size("Exit")

		while fontSurfSize1[0] > self.width or fontSurfSize2[0] > self.width or fontSurfSize3[0] > self.width\
		or fontSurfSize1[1] > self.height / 3 or fontSurfSize2[1] > self.height / 3 or fontSurfSize3[1] > self.height / 3:
			fontSize -= 1
			font = pygame.font.Font(pygame.font.get_default_font(), fontSize)
			fontSurfSize1 = font.size("Options")
			fontSurfSize2 = font.size("Pause")
			fontSurfSize3 = font.size("Exit")
						
		fontSize -= 1
		normalFont = pygame.font.Font(pygame.font.get_default_font(), fontSize)		
		
		self.optionsText = normalFont.render("Options", True, TEXT_NORMAL_COLOUR).convert_alpha()
		self.pauseText = normalFont.render("Pause", True, TEXT_NORMAL_COLOUR).convert_alpha()
		self.exitText = normalFont.render("Exit", True, TEXT_NORMAL_COLOUR).convert_alpha()
		
		highlightFont = pygame.font.Font(pygame.font.get_default_font(), fontSize)
		
		self.optionsTextHighlight = highlightFont.render("Options", True, TEXT_HIGHLIGHT_COLOUR).convert_alpha()
		self.pauseTextHighlight = normalFont.render("Pause", True, TEXT_HIGHLIGHT_COLOUR).convert_alpha()
		self.exitTextHighlight = highlightFont.render("Exit", True, TEXT_HIGHLIGHT_COLOUR).convert_alpha()
	
		self.surfRect = surf.get_rect(left = self.pos[0], top = self.pos[1])
		self.optionsRect = self.optionsText.get_rect(left = self.pos[0] + (self.width - self.optionsText.get_rect().width) / 2, top = self.pos[1] + ((self.height / 2) - self.optionsText.get_rect().height) / 2)
		self.pauseRect = self.pauseText.get_rect(left = self.pos[0] + (self.width - self.pauseText.get_rect().width) / 2, top = self.pos[1] + (2 * (self.height / 2) - self.pauseText.get_rect().height) / 2)
		self.exitRect = self.exitText.get_rect(left = self.pos[0] + (self.width - self.exitText.get_rect().width) / 2, top = self.pos[1] + (3 * (self.height / 2) - self.exitText.get_rect().height) / 2)
	
		self.bOptionsHigh = False
		self.bPauseHigh = False
		self.bExitHigh = False
		
		surf.blit(self.optionsText, ((self.width - self.optionsText.get_rect().width) / 2, ((self.height / 2) - self.optionsText.get_rect().height) / 2))
		surf.blit(self.pauseText, ((self.width - self.pauseText.get_rect().width) / 2, ((2 * self.height / 2) - self.pauseText.get_rect().height) / 2))
		surf.blit(self.exitText, ((self.width - self.exitText.get_rect().width) / 2, ((3 * self.height / 2) - self.exitText.get_rect().height) / 2))
		
		
		pygame.display.get_surface().blit(surf, self.pos)
		
		
	def updateMousePos(self, pos):	
		updateRects = []
		
		if self.surfRect.collidepoint(pos) or self.bOptionsHigh or self.bPauseHigh or self.bExitHigh:
			surf = pygame.Surface((self.width, self.height)).convert()
			surf.fill(BG_COLOUR)		
							
			self.bHighLight = False
					
			if self.optionsRect.collidepoint(pos):
				surf.blit(self.optionsTextHighlight, ((self.width - self.optionsTextHighlight.get_rect().width) / 2, ((self.height / 2) - self.optionsTextHighlight.get_rect().height) / 2))
				updateRects.append(self.optionsRect)
				
				self.bHighLight = True
				self.bOptionsHigh = True
			else:
				surf.blit(self.optionsText, ((self.width - self.optionsText.get_rect().width) / 2, ((self.height / 2) - self.optionsText.get_rect().height) / 2))
				updateRects.append(self.optionsRect)
				self.bOptionsHigh = False
		
			if self.pauseRect.collidepoint(pos) and not self.bHighLight:	
				surf.blit(self.pauseTextHighlight, ((self.width - self.pauseTextHighlight.get_rect().width) / 2, ((2 * self.height / 2) - self.pauseTextHighlight.get_rect().height) / 2))					
				updateRects.append(self.pauseRect)
				self.bHighLight = True
				self.bPauseHigh = True
			else:
				surf.blit(self.pauseText, ((self.width - self.pauseText.get_rect().width) / 2, ((2 * self.height / 2) - self.pauseText.get_rect().height) / 2))
				updateRects.append(self.pauseRect)
				self.bPauseHigh = False
		
			if self.exitRect.collidepoint(pos) and not self.bHighLight:	
				surf.blit(self.exitTextHighlight, ((self.width - self.exitTextHighlight.get_rect().width) / 2, ((3 * self.height / 2) - self.exitTextHighlight.get_rect().height) / 2))					
				updateRects.append(self.exitRect)	
				self.bHighLight = True
				self.bExitHigh = True
			else:
				surf.blit(self.exitText, ((self.width - self.exitText.get_rect().width) / 2, ((3 * self.height / 2) - self.exitText.get_rect().height) / 2))
				updateRects.append(self.exitRect)
				self.bExitHigh = False
				
			if updateRects:
				pygame.display.get_surface().blit(surf, self.pos)
		
		return updateRects
		
	def mouseClick(self, pos):
		if self.bOptionsHigh:
			pygame.event.post(pygame.event.Event(EVENT_OPTIONMENU, {}))
		elif self.bPauseHigh:
			pygame.event.post(pygame.event.Event(EVENT_PAUSE, {}))
		elif self.bExitHigh:
			sys.exit(1)
		

		
		
class GameArea:
	def __init__(self, world, space):
		# ODE
		self.space = space
		self.world = world
		self.contactgroup = ode.JointGroup()
		floor = ode.GeomPlane(space, (0, 1, 0), SCREEN_Y - (5 * GAME_SURFACE_SIZE[1] / 6))
		wall1 = ode.GeomPlane(space, (1, 0, 0), 0)
		wall2 = ode.GeomPlane(space, (-1, 0, 0), - (BORDER_WIDTH_X + GAME_SURFACE_SIZE[0]))
		
		self.walls = (wall1, wall2)
		
		#ode.GeomPlane(space, (0, 1, 0), 0)
				
		# Not ODE
		#self.bBlit = True
		
		self.baseSurface = pygame.image.load(filepath("background.jpg"))
		self.baseSurface = pygame.transform.scale(self.baseSurface, (GAME_SURFACE_SIZE[0], GAME_SURFACE_SIZE[1])).convert()
		self.cloudImage = pygame.image.load(filepath("cloud1.png"))	
		self.cloudImage = pygame.transform.scale(self.cloudImage, ((self.cloudImage.get_rect().width * SCREEN_X) / SCALETOSIZE[0], (self.cloudImage.get_rect().height * SCREEN_Y) / SCALETOSIZE[1])).convert_alpha()
			
		pygame.draw.line(self.baseSurface, TARGET_LINE_COLOUR, (BORDER_WIDTH_X, (1 - WIN_HEIGHT_MEDIUM) * SCREEN_Y), (BORDER_WIDTH_X + GAME_SURFACE_SIZE[0], (1 - WIN_HEIGHT_MEDIUM) * SCREEN_Y), 2)
			
		self.blockList = []
		self.cloudList = []
		
		while len(self.cloudList) < 8:
			defautlCloudRect = self.cloudImage.get_rect()
			scale = random.random() / 2 + 0.4
			self.cloudList.append(Cloud(pygame.transform.scale(self.cloudImage, (int(scale * defautlCloudRect.width), int(scale * defautlCloudRect.height))), True))

		
		self.windStrength = START_WIND_MEDIUM * ((-1) ** random.randint(0, 1))
		
		self.playableRect = pygame.Rect(0, BORDER_WIDTH_X, GAME_SURFACE_SIZE[0], 5 * GAME_SURFACE_SIZE[1] / 6)
		
		self.edgeRects = [(-100, BORDER_WIDTH_X + (5 * GAME_SURFACE_SIZE[1] / 6), GAME_SURFACE_SIZE[0] + 200, (GAME_SURFACE_SIZE[1] / 6) + 100), # Grass
						  (BORDER_WIDTH_X + GAME_SURFACE_SIZE[0], -100, SCREEN_X - BORDER_WIDTH_X + GAME_SURFACE_SIZE[0] + 100, SCREEN_Y + 200), # Right panel
						  (-100, -100, BORDER_WIDTH_X + 100, SCREEN_Y + 200), # Left border
						  (-100, -100, SCREEN_X + 200, BORDER_WIDTH_Y + 100), # Top border
						  (-100, SCREEN_Y - BORDER_WIDTH_Y, SCREEN_X + 200, BORDER_WIDTH_Y + 100)] # Bottom border
						  	
			
		#self.baseSurface = pygame.Surface(GAME_SURFACE_SIZE).convert()
				
		#self.baseSurface.fill((0, 0, 128))
		
		#self.baseSurface.fill((0, 128, 0), (0, 5 * GAME_SURFACE_SIZE[1] / 6, GAME_SURFACE_SIZE[0], 5 * GAME_SURFACE_SIZE[1] / 6))
		
		self.nextBlockPos = None
		
		pygame.display.get_surface().blit(self.baseSurface, (BORDER_WIDTH_X, BORDER_WIDTH_Y))
				
		self.gameSurface = self.baseSurface.copy()
				
	def snapToNearestValidPos(self, pos, block):
		#PERF - a tad slow
		#TODO - make collision detection based on ODE not pygame
	
		blockRect = block.getSurface().get_rect()
		blockRect.left -= blockRect.width / 2
		blockRect.top -= blockRect.height / 2
		
		blockRect.left += pos[0]
		blockRect.top += pos[1]
		
		
		rectNoCollideList = [] + self.edgeRects

		for newBlock in self.blockList:
			newBlockRect = newBlock.getSurface().get_rect()
			newBlockRect.topleft = newBlock.getPosition()
			
			rectNoCollideList.append(newBlockRect)
		
		testRect = blockRect
		
		bestPos = ()
		
		if blockRect.collidelist(rectNoCollideList) == -1:
			# Find nearest			
			for iIteration in range(1, max(GAME_SURFACE_SIZE)):		
				testRect = blockRect.move(0, iIteration)
				if testRect.collidelist(rectNoCollideList) != -1:
					bestPos = (testRect.left, testRect.top - 1)
					break
					
			blockRect.topleft = bestPos# = testRect
										
		if blockRect.collidelist(rectNoCollideList) != -1:
			# bump away
			bBreak = False
			for iIteration in range(1, max(GAME_SURFACE_SIZE) / 5):			
				bestDistSq = sys.maxint
								
				for i in range(-1 * iIteration, 2 * iIteration):
					for j in range(-1 * iIteration, 2 * iIteration, 2 * iIteration):
						testRect = blockRect.move(i, j)
						if testRect.collidelist(rectNoCollideList) == -1:
							if i == 0:
								bestPos = (testRect.left, testRect.top)
								bBreak = True
								break
						
							distSq = i ** 2 + j ** 2
							if distSq < bestDistSq:
								bestDistSq = distSq
								bestPos = (testRect.left, testRect.top)
									
					if bBreak:
						break
				if bBreak:
					break	
	
				for i in range(-1 * iIteration, 2 * iIteration, 2 * iIteration):
					for j in range(-1 * iIteration + 1, 2 * iIteration - 1):
						testRect = blockRect.move(i, j)
						if testRect.collidelist(rectNoCollideList) == -1:
							if j == 0:
								bestPos = (testRect.left, testRect.top)
								bBreak = True
								break							
						
							distSq = i ** 2 + j ** 2
							if distSq < bestDistSq:
								bestDistSq = distSq
								bestPos = (testRect.left, testRect.top)
								
					if bBreak:
						break								
									
				if bestPos:
					break
					
			blockRect.topleft = bestPos
			
		# Find nearest
		
		for iIteration in range(1, max(GAME_SURFACE_SIZE)):		
			testRect = blockRect.move(0, iIteration)
			if testRect.collidelist(rectNoCollideList) != -1:
				bestPos = (testRect.left, testRect.top - 1)
				break
					
			
		return bestPos

	
	def killBlock(self, pos):
		for block in self.blockList:
			if block.getRect(True).collidepoint(pos):
				block.quickKill()
		
	def updateMousePos(self, pos, currentBlock):	
		self.mousePos = pos
		self.mouseBlock = currentBlock
		
	def drawMouse(self):
		updateRectList = []
	
		if self.gameSurface.get_rect().collidepoint(self.mousePos):
			pygame.mouse.set_visible(False)

			blockSurf = self.mouseBlock.getSurface()
			blockRect = blockSurf.get_rect()
			
			self.nextBlockPos = newPos = self.snapToNearestValidPos(self.mousePos, self.mouseBlock)
					
			if newPos:
				updateRectList.append(self.gameSurface.blit(blockSurf, newPos))

			pygame.draw.circle(self.gameSurface, (255, 0, 0), self.mousePos, 7)
			updateRectList.append(pygame.draw.circle(self.gameSurface, (0, 0, 0), self.mousePos, 7, 1))
			
		else:
			pygame.mouse.set_visible(True)

			
		return updateRectList
		
	def setBlock(self, block, pos):
		
	
		if self.playableRect.collidepoint(pos):
			newPos = self.nextBlockPos
			if newPos:
				block.setInitialPosition(newPos)
				block.setSelected(False)
				self.blockList.append(block)
				
				#for block in self.blockList:
				#	self.gameSurface.blit(block.getSurface(), block.getPosition())
										
				#self.bBlit = True
				
				return True
			
		return False
		
	def doWind(self, dt):
		self.windStrength += (random.random() * MAX_WIND_CHANGE) - (MAX_WIND_CHANGE / 2.0)
		self.windStrength = clamp(self.windStrength, (-MAX_WIND_PRESSURE, MAX_WIND_PRESSURE))
		
		if len(self.cloudList) < 8:
			defautlCloudRect = self.cloudImage.get_rect()
			scale = random.random() / 2 + 0.4
			self.cloudList.append(Cloud(pygame.transform.scale(self.cloudImage, (scale * defautlCloudRect.width, scale * defautlCloudRect.height)), False))
		
		cloudRemoveList = []
	
		for cloud in self.cloudList:
			if cloud.move(self.windStrength, dt):
				cloudRemoveList.append(cloud)
				
		for cloud in cloudRemoveList:
			self.cloudList.remove(cloud)
	
		for block in self.blockList:
			# +ve means in +ve x, so coming from -ve x
			rect1 = block.getRect()
			rect1.topleft = block.getPosition()
			
			windMod = 1
			for block2 in self.blockList:
				if block != block2:
					rect2 = block2.getRect()
					rect2.topleft = block2.getPosition()
					
					#print rect1, rect2
					
					if rect1.bottom < rect2.top:
						continue
					elif rect1.top > rect2.bottom:
						continue
					else:
						if self.windStrength > 0:
							if rect1.left < rect2.right:
								continue
							distance = rect1.right - rect2.left
						else:
							if rect1.right > rect2.left:
								continue
							distance = rect2.right - rect1.left
														
					if distance > SCREEN_Y / 4:
						continue
							
					# Consider changing this to a discreet system involving wind effective points?
					if rect1.top > rect2.top:
						if rect1.bottom > rect2.bottom:
							breakLength = rect2.bottom - rect1.top
						else:
							breakLength = rect1.height
					elif rect1.top <= rect2.top:
						if rect1.bottom >= rect2.bottom:
							breakLength = rect2.height
						else:
							breakLength = rect1.bottom - rect2.top					
					
					if not breakLength:
						continue
						
					
					if breakLength > rect1.height:
						crash
					
					distanceFraction = math.sqrt(distance / float(SCREEN_Y)) / 2
					breakFraction = (1 - (breakLength / float(rect1.height))) / 2
					
					windMod *= (distanceFraction + breakFraction)
					
					
			block.doWindForces(self.windStrength * dt * windMod)
			
		
	def checkVictoryConditions(self, dt):
		maxVicTimer = 0
		
		for block in self.blockList:
			if block.getPosition()[1] < (1 - WIN_HEIGHT_MEDIUM) * SCREEN_Y:
				vicTime = block.incVictoryTimer(dt)
				if vicTime > maxVicTimer:
					maxVicTimer = vicTime
			else:
				block.resetVictoryTimer()
				
		return maxVicTimer
			
	def update(self, dt, bNoMouse = False):		
		updateRectList = []
	
		self.gameSurface = self.baseSurface.copy()
	
		if dt > 0:
			self.doWind(dt)
			
		for cloud in self.cloudList:		
			updateRectList.append(self.gameSurface.blit(cloud.getSurface(), cloud.getPosition()))
			
		if dt > 0:
			n = 5
			for i in range(n):
				self.space.collide((self.world, self.contactgroup, self.blockList, self.walls), near_callback)
				self.world.quickStep(dt / n)
				self.contactgroup.empty()
			
			blockKillList = []
			
		for block in self.blockList:
			if dt > 0:	
				if block.update():
					blockKillList.append(block)
					continue
			
				rect = block.getSurface().get_rect()
				bodyPos = block.getBody().getPosition()

				block.setPosition(((bodyPos[0] - rect.center[0]), (SCREEN_Y - (bodyPos[1] + rect.center[1]))))
			
			updateRectList.append(self.gameSurface.blit(block.getSurface(), block.getPosition()))
			
		if dt > 0:	
			for block in blockKillList:
				self.blockList.remove(block)
			
		if dt > 0 and not bNoMouse:	
			updateRectList += self.drawMouse()
	
		pygame.display.get_surface().blit(self.gameSurface, (BORDER_WIDTH_X, BORDER_WIDTH_Y))
						
		return updateRectList
			
						
			
class Block:
	def __init__(self, world, space):
		if not random.randint(0, 5):
			self.height = random.randint(1, 6) * (MAX_BLOCK_SIZE_Y / 6)
		elif not random.randint(0, 3):
			self.height = random.randint(1, 3) * (MAX_BLOCK_SIZE_Y / 6)
		elif not random.randint(0, 1):
			self.height = random.randint(1, 2) * (MAX_BLOCK_SIZE_Y / 6)
		else:
			self.height = MAX_BLOCK_SIZE_Y / 6
						
		if not random.randint(0, 5):
			self.width = random.randint(1, 6) * (MAX_BLOCK_SIZE_X / 6)
		elif not random.randint(0, 3):
			self.width = random.randint(1, 3) * (MAX_BLOCK_SIZE_X / 6)
		elif not random.randint(0, 1):
			self.width = random.randint(1, 2) * (MAX_BLOCK_SIZE_X / 6)
		else:	  
			self.width = MAX_BLOCK_SIZE_X / 6
			
		if random.randint(0, 1):
			if random.randint(0, 1):
				self.height = min(MAX_BLOCK_SIZE_Y, self.height * 2)
			else:
				self.width = min(MAX_BLOCK_SIZE_X, self.width * 2)
							 		
		self.surf = pygame.Surface((self.width, self.height)).convert()
		self.surf.set_colorkey((0, 255, 0))
		
		self.surf.fill((0, 0, 0))
		self.surf.fill(OLD_BLOCK_COLOUR, (1, 1, self.width - 2, self.height - 2))
		
		self.pos = [0, 0]
		# self.oldPositions = [(-1, -1) for i in range(5)]
		# self.iOldPosLoc = 0		
		
		# self.oldBodyRotation = None
		# self.oldRectRotation = None
		# self.oldBodyPosition = None
		
		# self.oldRect = (0,0,0,0)
		
		self.world = world
		self.space = space
		
		self.body = None
		
		self.bSelected = True
		self.iVictoryTimer = 0
		
		self.iKillTimer = -BLOCK_KILL_NUM_ROTATION_TICKS
		self.bQuickKill = False
		self.bDead = False

	def kill(self):
		if self.iKillTimer < 0:
			self.iKillTimer += 1
			
	def quickKill(self):
		if not self.bQuickKill:
			self.iKillTimer = max(0, self.iKillTimer)
			self.bQuickKill = True
			
	def incVictoryTimer(self, dt):
		self.iVictoryTimer += dt
		return self.iVictoryTimer
		
	def resetVictoryTimer(self):
		self.iVictoryTimer = 0

	def getBody(self):
		return self.body
		
	def getRect(self, bIncldePos = False):
		if bIncldePos:		
			rect = self.getSurface().get_rect()
			pos = self.getPosition()
			rect.left += pos[0]
			rect.top += pos[1]
			
			return rect
		else:
			return self.getSurface().get_rect()
			
	def getOldRect(self):
		return self.oldRect
		
	def getSurface(self):
		if self.body:
			rotation = self.getRotation()
			
			if int(rotation) % 90 > 10 and int(rotation) % 90 < 80:
				self.kill()
							
			surf = pygame.transform.rotate(self.surf, rotation)
		else:
			surf = self.surf
			
	
		return surf
		
	def getPosition(self):
		return self.pos
		
	def getRotation(self):
		bodyRotation = self.body.getRotation()

		rotation = math.atan2(clamp(bodyRotation[1], (-1, 1)), clamp(bodyRotation[0], (-1, 1)))
		
		self.body.setRotation([math.cos(rotation), math.sin(rotation), 0,
							   -math.sin(rotation), math.cos(rotation), 0,
							   0, 0, 1])
		
		rotation *= 180 / math.pi
					
		rotation = -rotation		
		
		return rotation

	def setSelected(self, bSelected):
		self.bSelected = bSelected
		if not bSelected:
			self.surf.fill((0, 0, 0))
			self.surf.fill(OLD_BLOCK_COLOUR, (1, 1, self.width - 2, self.height - 2))	
		else:
			self.surf.fill((0, 0, 0))
			self.surf.fill(SELECTED_BLOCK_COLOUR, (1, 1, self.width - 2, self.height - 2))
		
	def setInitialPosition(self, pos):
		rect = self.surf.get_rect()
	
		# Create body
		self.body = ode.Body(self.world)
		self.body.setPosition(((pos[0] + rect.center[0]), (SCREEN_Y - (pos[1] + rect.center[1])), 0))
		M = ode.Mass()
		M.setBox(BLOCK_DENSITY / 500.0, self.width, self.height, 500)
		self.body.setMass(M)

		# Create a box geom for collision detection
		self.geom = ode.GeomBox(self.space, lengths = (self.width, self.height, 500))
		self.geom.setBody(self.body)	
		
		self.pos = pos
		# self.oldBodyRotation = self.body.getRotation()
		# self.oldRectRotation = self.getRotation()
		# self.oldBodyPosition = self.body.getPosition()
		
	def setPosition(self, pos):	
		#self.oldPositions[self.iOldPosLoc] = pos
				
		# avgX = 0.0
		# avgY = 0.0
				
		# numEntries = 0
				
		# for oldPos in self.oldPositions:
			# if oldPos[0] > -1 and oldPos[1] > -1:
				# avgX += oldPos[0]
				# avgY += oldPos[1]
				
				# numEntries += 1
				
		#HACK
		# if numEntries > 0:
			# avgX /= numEntries
			# avgY /= numEntries
		# else:
			# avgX = -1
			# avgY = -1
					
		#if True or math.fabs(pos[0] - avgX) > 0.5 or math.fabs(pos[1] - avgY) > 0.5 or math.fabs(self.getRotation() - self.oldRectRotation) > 0.1:
		self.pos = (pos[0], pos[1])
			# self.oldBodyRotation = self.body.getRotation()
			# self.oldRectRotation = self.getRotation()
			# self.oldBodyPosition = self.body.getPosition()
		# else:
			# rect = self.surf.get_rect()
			# self.body.setPosition(self.oldBodyPosition)
			# self.body.setRotation(self.oldBodyRotation)
		
		#self.iOldPosLoc = (self.iOldPosLoc % 4) + 1
		
	def doWindForces(self, windStrength):
		if not windStrength:
			return
		windStrength *= 40000000
	
		rect = self.getRect()
		rect.topleft = self.getPosition()
		
		windLength = rect.height / float(NUM_WIND_POINTS)
			
		windEffStr = windStrength * windLength
		
		for i in range(NUM_WIND_POINTS):
			height = ((((i * 2) + 1) / float((NUM_WIND_POINTS * 2) + 1)) * rect.height) + rect.top	
						
			#print 1 - (math.sqrt(height / float(SCREEN_Y)))
			
			#print height, float(SCREEN_Y)
			
			if height >= 0:
				self.getBody().addForceAtPos((windEffStr * (1 - (math.sqrt(height / float(SCREEN_Y)))), 0, 0), (rect.center[0] + self.getPosition()[0], SCREEN_Y - height, 0))
			else:
				self.getBody().addForceAtPos((windEffStr, 0, 0), (rect.center[0] + self.getPosition()[0], SCREEN_Y - height, 0))
		
	def update(self):
		if self.iKillTimer >= BLOCK_KILL_TIME:
			self.body.disable()
			self.space.remove(self.geom)
			return True
			
		if self.iKillTimer >= BLOCK_KILL_TIME / 2:
			if not self.bDead:
				self.bDead = True

			
		if self.iKillTimer >= 0:

			if self.bQuickKill:
				self.iKillTimer += 5 
			else:
				self.iKillTimer += 1 
				
				
			#self.surf.convert()
			#TODO - try per surface alphas. Didn't work before, but worth a try. REALLY QUITE IMPORTANT TO PERFORMANCE
			#self.surf = self.surf.conver()
			#self.surf.set_alpha(255 - 255 * (self.iKillTimer / BLOCK_KILL_TIME) ** 2)
			
			
			self.surf = self.surf.convert_alpha()
			self.surf.fill((0, 0, 0, 0))		
			
			self.surf.fill((0, 0, 0, int(255 - 255 * min(1, self.iKillTimer / BLOCK_KILL_TIME) ** 2)))
			self.surf.fill((0, 0, 192, 255 - int(255 * min(1, self.iKillTimer / BLOCK_KILL_TIME) ** 2)), (1, 1, self.width - 2, self.height - 2))

				
		return False
		
		
class Cloud:
	def __init__(self, surface, bFirst):
		self.surface = surface
		
		
		if bFirst:
			self.position = [BORDER_WIDTH_X - self.surface.get_rect().width + int(random.random() * (GAME_SURFACE_SIZE[0] + self.surface.get_rect().width)), int(random.random() * SCREEN_Y) / 4]
		else:
			if random.random() > 0.5:
				self.position = [BORDER_WIDTH_X - self.surface.get_rect().width, int(random.random() * SCREEN_Y) / 4]
			else:
				self.position = [BORDER_WIDTH_X + GAME_SURFACE_SIZE[0], int(random.random() * SCREEN_Y) / 4]
				
		self.cloudSlowness = (random.random() + 0.5) * 0.75
			
		#self.position = [100, 20]
		
	def move(self, force, timeStep):
		self.position[0] += force * timeStep / self.cloudSlowness 
		if self.position[0] < BORDER_WIDTH_X - self.surface.get_rect().width or self.position[0] >  BORDER_WIDTH_X + GAME_SURFACE_SIZE[0]:
			return True
		return False
		
	def getSurface(self):
		return self.surface		
		
	def getPosition(self):
		return self.position
		
		
def clamp(value, limits):
	if value < limits[0]:
		value = limits[0]
	if value > limits[1]:
		value = limits[1]
		
	return value
		
# Collision callback
def near_callback(args, geom1, geom2):
	world, contactgroup, blockList, walls = args

	# Check if the objects do collide
	contacts = ode.collide(geom1, geom2)
	
	if contacts:
		if geom1 in walls:
			for block in blockList:
				if block.geom == geom2:
					block.quickKill()
		elif geom2 in walls:
			for block in blockList:
				if block.geom == geom1:
					block.quickKill()				
		else:
			block1 = None
			block2 = None
			
			bOnlyGeom1 = False
			bOnlyGeom2 = False
			
			for block in blockList:
				if block.geom == geom1:
					block1 = block
				elif block.geom == geom2:
					block2 = block
					
				if block1 and block2:
					break
					
			if block1:
				if block1.bDead:
					bOnlyGeom1 = True
			if block2:
				if block2.bDead:
					bOnlyGeom2 = True				
					
	
		# Create contact joints
		for c in contacts:
			c.setBounce(0.1)
			c.setBounceVel(200)
			c.setMu(250000000000)
			#c.setMu(250)
			j = ode.ContactJoint(world, contactgroup, c)
			j.attach(geom1.getBody(), geom2.getBody())	
		
if __name__ == "__main__":
	pygame.init()
	
	import profile	
	profile.run("runGame()")
		
		